#include "stdafx.h"
#include "textfile.h"
#include "ktlarray.h"
using namespace ki;



//=========================================================================
// eLXgt@CǂݏoʃC^[tFCX
//=========================================================================

struct ki::TextFileRPimpl : public Object
{
	inline TextFileRPimpl()
		: state(EOL) {}

	virtual ulong ReadLine( unicode* buf, ulong siz ) = 0;

	enum { EOF=0, EOL=1, EOB=2 } state;
};



//-------------------------------------------------------------------------
// Unicodenp̃x[XNX
//	UTF-8ȊO͂ȂɏoȂ낤xĂ悵ƂB
//-------------------------------------------------------------------------

struct BasicUTF : public ki::TextFileRPimpl
{
	virtual unicode PeekC() = 0;
	virtual unicode GetC() {unicode ch=PeekC(); Skip(); return ch;}
	virtual void Skip() = 0;
	virtual bool Eof() = 0;

	ulong ReadLine( unicode* buf, ulong siz )
	{
		state = EOF;

		// so܂œǂ
		unicode *w=buf, *e=buf+siz;
		while( !Eof() )
		{
			*w = GetC();
			if( *w==L'\r' || *w==L'\n' )
			{
				state = EOL;
				break;
			}
			else if( *w!=0xfeff && ++w==e )
			{
				state = EOB;
				break;
			}
		}

		// sR[hXLbv
		if( state == EOL )
			if( *w==L'\r' && !Eof() && PeekC()==L'\n' )
				Skip();

		// ǂ񂾕
		return w-buf;
	}
};



//-------------------------------------------------------------------------
// UCS2x^/UCS4x^BꂼUTF16, UTF32̑ƂĎgB
// łɓtemplateŁAISO-8859-1Ă܂B^^;
//-------------------------------------------------------------------------

template<typename T>
struct UCS : public BasicUTF
{
	UCS( const uchar* b, ulong s, bool bigendian )
		: fb( reinterpret_cast<const T*>(b) )
		, fe( reinterpret_cast<const T*>(b+(s/sizeof(T))*sizeof(T)) )
		, be( bigendian ) {}

	const T *fb, *fe;
	const bool be;

	// GfBAϊ
	inline  byte swap(  byte val ) { return val; }
	inline dbyte swap( dbyte val ) { return (val<<8) |(val>>8); }
	inline qbyte swap( qbyte val ) { return ((val>>24)|(val>>8)); }

	void Skip() { ++fb; }
	bool Eof() { return fb==fe; }
	unicode PeekC() { return (unicode)(be ? swap(*fb) : *fb); }
};

typedef UCS< byte> West;
typedef UCS<dbyte> Utf16;
typedef UCS<qbyte> Utf32;



//-------------------------------------------------------------------------
// UTF-5
//     0-  F : 1bbbb
//    10- FF : 1bbbb 0bbbb
//   100-FFF : 1bbbb 0bbbb 0bbbb
// Ƃ悤ɁA16ił̈ꌅꕶŕ\ĂtH[}bgB
// e 0bbbb  '0', '1', ... '9', 'A', ... 'F'
// e 1bbbb  'G', 'H', ... 'P', 'Q', ... 'V' ̎ŕ\B
//-------------------------------------------------------------------------

struct Utf5 : public BasicUTF
{
	Utf5( const uchar* b, ulong s )
		: fb( b )
		, fe( b+s ) {}

	const uchar *fb, *fe;

	// 16i琮l֕ϊ
	inline byte conv( uchar x )
	{
		if( '0'<=x && x<='9' ) return x-'0';
		else                   return x-'A'+0x0A;
	}

	void Skip() { do ++fb; while( fb<fe && *fb<'G' ); }
	bool Eof() { return fb==fe; }
	unicode PeekC()
	{
		unicode ch = (*fb-'G');
		for( const uchar* p=fb+1; p<fe && *p<'G'; ++p )
			ch = (ch<<4)|conv(*p);
		return ch;
	}
};



//-------------------------------------------------------------------------
// UTF-7
//   ASCII͈͂̎͂̂܂܁BȊOUTF-16̒lbase64GR[h
//   ďóBGR[hꂽ +  - ŋ܂B܂ '+' 
//   ̂\邽߂ "+-" Ƃ`pB
//-------------------------------------------------------------------------

namespace
{
	static const uchar u7c[128]={
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x3e,0xff,0xff,0xff,0x3f,
	0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,
	0x0f,0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0xff,0xff,0xff,0xff,0xff,
	0xff,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f,0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,
	0x29,0x2a,0x2b,0x2c,0x2d,0x2e,0x2f,0x30,0x31,0x32,0x33,0xff,0xff,0xff,0xff,0xff };
}

struct Utf7 : public BasicUTF
{
	Utf7( const uchar* b, ulong s )
		: fb( b )
		, fe( b+s )
		, rest( -1 ) { fillbuf(); }

	const uchar *fb, *fe;
	unicode buf[3]; // b64WɓǂŃobt@ɗ߂Ă
	int rest;       // obt@̋
	bool inB64;     // base64GAȂtrue

	void Skip() { if(--rest==0) fillbuf(); }
	bool Eof() { return fb==fe && rest==0; }
	unicode PeekC() { return buf[rest-1]; }

	void fillbuf()
	{
		if( fb<fe )
			if( !inB64 )
				if( *fb=='+' )
					if( fb+1<fe && fb[1]=='-' )
						rest=1, buf[0]=L'+', fb+=2;  // +-
					else
						++fb, inB64=true, fillbuf(); // P +
				else
					rest=1, buf[0]=*fb++;            // ʂ̎
			else
			{
				// fR[hł邩
				int N=0, E=Max(fb-fe,8);
				while( N<E && fb[N]<0x80 && u7c[fb[N]]!=0xff )
					++N;

				// fR[h
				buf[0]=buf[1]=buf[2]=0;
				switch( N )
				{
				case 8: buf[2]|= u7c[fb[7]];
				case 7: buf[2]|=(u7c[fb[6]]<< 6);
				case 6: buf[2]|=(u7c[fb[5]]<<12), buf[1]|=(u7c[fb[5]]>>4);
				case 5: buf[1]|=(u7c[fb[4]]<< 2);
				case 4: buf[1]|=(u7c[fb[3]]<< 8);
				case 3: buf[1]|=(u7c[fb[2]]<<14), buf[0]|=(u7c[fb[2]]>>2);
				case 2: buf[0]|=(u7c[fb[1]]<< 4);
				case 1: buf[0]|=(u7c[fb[0]]<<10);
					unicode t;
					rest = 1;
					if( N==8 )
						rest=3, t=buf[0], buf[0]=buf[2], buf[2]=t;
					else if( N>=6 )
						rest=2, t=buf[0], buf[0]=buf[1], buf[1]=t;
				}

				// gi
				if( N<E )
				{
					inB64=false;
					if( fb[N]=='-' )
						++fb;
				}
				fb += N;
				if( N==0 )
					fillbuf();
			}
	}
};



//-------------------------------------------------------------------------
// UTF8/MBCS
//  CR,LFPoCgƂĂƏoĂ̂ŁA
//  ؂蕪ȒPȌ`ł܂Ƃ߂ĈBUTF8ȊO̕ϊ
//  WindowsɑSĔCĂB
//-------------------------------------------------------------------------

namespace 
{
	typedef int (WINAPI * uConvFunc)(UINT,DWORD,const char*,int,wchar_t*,int);
	typedef char* (WINAPI * uNextFunc)(WORD,const char*,DWORD);

	static const byte mask[] = { 0, 0xff, 0x1f, 0x0f, 0x07, 0x03, 0x01 };
	static inline int GetMaskIndex(uchar n)
	{
		if( uchar(n+2) < 0xc2 ) return 1; // 00`10111111, fe, ff
		if( n          < 0xe0 ) return 2; // 110xxxxx
		if( n          < 0xf0 ) return 3; // 1110xxxx
		if( n          < 0xf8 ) return 4; // 11110xxx
		if( n          < 0xfc ) return 5; // 111110xx
		                        return 6; // 1111110x
	}

	static int WINAPI Utf8ToWideChar( UINT, DWORD, const char* sb, int ss, wchar_t* wb, int ws )
	{
		const uchar *p = reinterpret_cast<const uchar*>(sb);
		const uchar *e = reinterpret_cast<const uchar*>(sb+ss);
		wchar_t     *w = wb; // obt@TCY`FbNidlj

		for( int t; p<e; ++w )
		{
			t  = GetMaskIndex(*p);
			*w = (*p++ & mask[t]);
			while( p<e && --t )
				*w<<=6, *w|=(*p++)&0x3f;
		}
		return w-wb;
	}

	static char* WINAPI CharNextUtf8( WORD, const char* p, DWORD )
	{
		return const_cast<char*>( p+GetMaskIndex(uchar(*p)) );
	}
}

struct MBCS : public TextFileRPimpl
{
	// t@C|C^R[hy[W
	const char* fb;
	const char* fe;
	const int   cp;
	uConvFunc conv;
	uNextFunc next;

	// ݒ
	MBCS( const uchar* b, ulong s, int c )
		: fb( reinterpret_cast<const char*>(b) )
		, fe( reinterpret_cast<const char*>(b+s) )
		, cp( c )
		, conv( cp==UTF8 ? Utf8ToWideChar : MultiByteToWideChar )
		, next( cp==UTF8 ? CharNextUtf8   : CharNextExA         )
	{
		if( cp==UTF8 && fe-fb>=3
		 && b[0]==0xef && b[1]==0xbb && b[2]==0xbf )
			fb += 3; // BOMXLbv
	}

	ulong ReadLine( unicode* buf, ulong siz )
	{
		// obt@̏I[At@C̏I[̋߂܂œǂݍ
		const char *p, *end = Min( fb+siz/2, fe );
		state = (end==fe ? EOF : EOB);

		// so܂Ői
		for( p=fb; p<end; )
			if( *p=='\r' || *p=='\n' )
			{
				state = EOL;
				break;
			}
			else if( (*p) & 0x80 )
			{
				p = next(cp,p,0);
			}
			else
			{
				++p;
			}

		// Unicode֕ϊ
		ulong len = conv( cp, MB_PRECOMPOSED, fb, p-fb, buf, siz );

		// sR[hXLbv
		if( state == EOL )
			if( *(p++)=='\r' && p<fe && *p=='\n' )
				++p;
		fb = p;

		// I
		return len;
	}
};



//-------------------------------------------------------------------------
// ISO-2022 ̓KȎ
//
//	R[hy[WƂāAG0, G1, G2, G3 ̎lʂB
//	ꂼꌈ܂GXP[vV[PXɂāA
//	܂܂ȕWeʂɌĂяoƂoB
//	Ƃ肠݂̃o[WőΉĂӂȂ̂
//	̒ʂB()ɁÂՂB
//
//		<<ł>>
//		1B 28 42    : G0 ASCII
//		1B 28 4A    : G0 JIS X 0201 [} (̂ASCII)
//		1B 29 4A    : G1 JIS X 0201 [} (̂ASCII)
//		1B 2A 4A    : G2 JIS X 0201 [} (̂ASCII)
//		1B 3B 4A    : G3 JIS X 0201 [} (̂ASCII)
//		1B 2E 41    : G2 ISO-8859-1
//		<<CP932Lȏꍇ>>
//		1B 28 49    : G0 JIS X 0201 Ji
//		1B 29 49    : G1 JIS X 0201 Ji
//		1B 2A 49    : G2 JIS X 0201 Ji
//		1B 2B 49    : G3 JIS X 0201 Ji
//		1B 24 40    : G0 JIS X 0208(1978)
//		1B 24 42    : G0 JIS X 0208(1983)    (Nx͋ʂȂ)
//		<<CP936Lȏꍇ>>
//		1B 24 41    : G0 GB 2312
//		1B 24 29 41 : G1 GB 2312
//		1B 24 2A 41 : G2 GB 2312
//		1B 24 2B 41 : G3 GB 2312
//		<<CP949Lȏꍇ>>
//		1B 24 28 43 : G0 KS X 1001
//		1B 24 29 43 : G1 KS X 1001
//		1B 24 2A 43 : G2 KS X 1001
//		1B 24 2B 43 : G3 KS X 1001
//
//	eʂɌĂяoẂA
//		GL (0x21`0xfe) GR (0xa0`0xff)
//	̂ǂ炩փ}bv邱ƂŁAۂ̃oCglƂȂB
//	}bv߂ƂȂoCǵA̒ʂ
//
//		0F    : GL        G0Ăяo
//		0E    : GL        G1Ăяo
//		1B 7E : GR        G1Ăяo
//		8E    : GL/GR G2uĂяoB1B 4E `
//		8F    : GL/GR G3uĂяoB1B 4F `
//
//-------------------------------------------------------------------------

enum CodeSet { ASCII, LATIN, KANA, JIS, KSX, GB };

struct Iso2022 : public TextFileRPimpl
{
	// Helper: JIS X 0208 => SJIS
	void jis2sjis( uchar k, uchar t, char* s )
	{
		if(k>=0x3f)	s[0] = (char)(((k+1)>>1)+0xc0);
		else		s[0] = (char)(((k+1)>>1)+0x80);
		if( k&1 )	s[1] = (char)((t>>6) ? t+0x40 : t+0x3f);
		else		s[1] = (char)(t+0x9e);
	}

	// t@C|C^
	const uchar* fb;
	const uchar* fe;
	bool      fixed; // ESCɂ؂ւsȂȂtrue
	bool    mode_hz; // HZ̏ꍇB

	// ƕϐ
	CodeSet *GL, *GR, G[4];
	int gWhat; // ̎ 1:GL/GR 2:G2 3:G3 ŏo
	ulong len;

	// 
	Iso2022( const uchar* b, ulong s, bool f, bool hz,
		CodeSet g0, CodeSet g1, CodeSet g2=ASCII, CodeSet g3=ASCII )
		: fb( b )
		, fe( b+s )
		, fixed( f )
		, mode_hz( hz )
		, GL( &G[0] )
		, GR( &G[1] )
		, gWhat( 1 )
	{
		G[0]=g0, G[1]=g1, G[2]=g2, G[3]=g3;
	}

	void DoSwitching( const uchar*& p )
	{
		if( p[1]==0x4A )
			G[ (p[0]-0x28)%4 ] = ASCII;         // 1B [28-2B] 4A
		else if( p[1]==0x49 )
			G[ (p[0]-0x28)%4 ] = KANA;          // 1B [28-2B] 49
		else if( *reinterpret_cast<const dbyte*>(p)==0x4228 )
			G[ 0 ] = ASCII;                     // 1B 28 42
		else if( *reinterpret_cast<const dbyte*>(p)==0x412E )
			G[ 2 ] = LATIN;                     // 1B 2E 41
		else if( p[0]==0x24 )
			if( p[1]==0x40 || p[1]==0x42 )
				G[ 0 ] = JIS;                   // 1B 24 [40|42]
			else if( p[1]==0x41 )
				G[ 0 ] = GB;                    // 1B 24 41
			else if( p+2 < fe )
				if( p[2]==0x41 )
					G[ ((*++p)-0x28)%4 ] = GB;  // 1B 24 [28-2B] 41
				else if( p[2]==0x43 )
					G[ ((*++p)-0x28)%4 ] = KSX; // 1B 24 [28-2B] 43
		++p;
	}

	void DoOutput( unicode*& buf, const uchar*& p )
	{
		// Wo
		CodeSet cs =
			(gWhat==2 ? G[2] : 
			(gWhat==3 ? G[3] :
			(*p&0x80  ? *GR  : *GL)));

		char c[2];
		ulong wt=1;
		switch( cs )
		{
		case ASCII:
			*buf = (*p)&0x7f;
			break;
		case LATIN:
			*buf = (*p)|0x80;
			break;
		case KANA:
			c[0] = (*p)|0x80;
			wt = ::MultiByteToWideChar(
				932, MB_PRECOMPOSED, c, 1, buf, 2 );
			break;
		case GB:
		case KSX:
			c[0] = (*  p)|0x80;
			c[1] = (*++p)|0x80;
			wt = ::MultiByteToWideChar(
				(cs==GB?936:949), MB_PRECOMPOSED, c, 2, buf, 2 );
			break;
		case JIS:
			jis2sjis( (p[0]&0x7f)-0x20, (p[1]&0x7f)-0x20, c );
			++p;
			wt = ::MultiByteToWideChar(
				932, MB_PRECOMPOSED, c, 2, buf, 2 );
			break;
		}
		buf+=wt;
		len+=wt;
	}

	ulong ReadLine( unicode* buf, ulong siz )
	{
		len=0;

		// obt@̏I[At@C̏I[̋߂܂œǂݍ
		const uchar *p, *end = Min( fb+siz/2, fe );
		state = (end==fe ? EOF : EOB);

		// so܂Ői
		for( p=fb; p<end; ++p )
			switch( *p )
			{
			case '\r':
			case '\n': state =   EOL; goto outofloop;
			case 0x0F:    GL = &G[0]; break;
			case 0x0E:    GL = &G[1]; break;
			case 0x8E: gWhat =     2; break;
			case 0x8F: gWhat =     3; break;
			case 0x1B:
				if( p+1<fe ) {
					++p; if( *p==0x7E )         GR = &G[1];
					else if( *p==0x4E )         gWhat =  2;
					else if( *p==0x4F )         gWhat =  3;
					else if( p+1<fe && !fixed ) DoSwitching(p);
				}break;
			case 0x7E: if( mode_hz && p+1<fe ) {
					++p; if( *p==0x7D ){ GL = &G[0]; break; }
					else if( *p==0x7B ){ GL = &G[1]; break; }
				} // fall through...
			default:   DoOutput( buf, p ); gWhat=1; break;
			}
		outofloop:

		// sR[hXLbv
		if( state == EOL )
			if( *(p++)=='\r' && p<fe && *p=='\n' )
				++p;
		fb = p;

		// I
		return len;
	}
};



//-------------------------------------------------------------------------
// ȂǂȂ
//-------------------------------------------------------------------------

TextFileR::TextFileR( int charset )
	: cs_( charset )
{
	// Nothing to do, but it's ok.
}

TextFileR::~TextFileR()
{
	// Pimpl̃fXgN^Ăяop
}

ulong TextFileR::ReadLine( unicode* buf, ulong siz )
{
	// 񂷂
	return impl_->ReadLine( buf, siz );
}

int TextFileR::state() const
{
	// 񂷂
	return impl_->state;
}

void TextFileR::Close()
{
	// 邾
	fp_.Close();
}

bool TextFileR::Open( const TCHAR* fname )
{
	// t@CJ
	if( !fp_.Open(fname) )
		return false;
	const uchar* buf = fp_.base();
	const ulong  siz = fp_.size();

	// KvȂ玩
	cs_ = AutoDetection( cs_, buf, Min<ulong>(siz,16<<10) ); // 擪16KB

	// ΉfR[_쐬
	switch( cs_ )
	{
	case Western: impl_ = new West(buf,siz,true);                         break;
	case UTF16b:
	case UTF16BE: impl_ = new Utf16(buf,siz,true);                        break;
	case UTF16l:
	case UTF16LE: impl_ = new Utf16(buf,siz,false);                       break;
	case UTF32b:
	case UTF32BE: impl_ = new Utf32(buf,siz,true);                        break;
	case UTF32l:
	case UTF32LE: impl_ = new Utf32(buf,siz,false);                       break;
	case UTF5:    impl_ = new Utf5(buf,siz);                              break;
	case UTF7:    impl_ = new Utf7(buf,siz);                              break;
	case EucJP:   impl_ = new Iso2022(buf,siz,true,false,ASCII,JIS,KANA); break;
	case IsoJP:   impl_ = new Iso2022(buf,siz,false,false,ASCII,KANA);    break;
	case IsoKR:   impl_ = new Iso2022(buf,siz,true,false,ASCII,KSX);      break;
	case IsoCN:   impl_ = new Iso2022(buf,siz,true,false,ASCII,GB);       break;
	case HZ:      impl_ = new Iso2022(buf,siz,true,true, ASCII,GB);       break;
	default:      impl_ = new MBCS(buf,siz,cs_);                          break;
	}

	return true;
}

int TextFileR::AutoDetection( int cs, const uchar* ptr, ulong siz )
{
//-- ܂Ȁo񐔂̓v

	int freq[256];
	bool bit8 = false;
	memzero( freq, sizeof(freq) );
	for( ulong i=0; i<siz; ++i )
	{
		if( ptr[i] >= 0x80 )
			bit8 = true;
		++freq[ ptr[i] ];
	}

//-- sR[h (UTF16,UTF32̂Ƃ肠BUTF5ɎĂ͔sc)

	     if( freq['\r'] > freq['\n']*2 ) lb_ = 0; // CR
	else if( freq['\n'] > freq['\r']*2 ) lb_ = 1; // LF
	else                                 lb_ = 2; // CRLF炢

//-- ȂƂʂꍇ͂ŏI

	if( cs != AutoDetect )
		return cs;

	switch( ::GetACP() )
	{
	case 932: cs = SJIS;    break;
	case 936: cs = GBK;     break;
	case 949: cs = UHC;     break;
	case 950: cs = Big5;    break;
	default:  cs = Western; break;
	}

	if( siz <= 16 )
		return cs;

//-- BOM`FbNE7bit`FbN

	bool Jp = ::IsValidCodePage(932)!=FALSE;
	bool Cn = ::IsValidCodePage(936)!=FALSE;
	bool Kr = ::IsValidCodePage(949)!=FALSE;

	if( ptr[0]==0x00 && ptr[1]==0x00 && ptr[2]==0xfe && ptr[3]==0xff )
		return UTF32b;
	if( ptr[0]==0xef && ptr[1]==0xbb && ptr[2]==0xbf )
		return UTF8;
	if( ptr[0]==0xfe && ptr[1]==0xff )
		return UTF16b;
	if( ptr[0]==0xff && ptr[1]==0xfe )
		return ( ptr[2]==0x00 && ptr[3]==0x00 ) ? UTF32l : UTF16l;
	if( Kr && ptr[0]==0x1b && ptr[1]==24 && ptr[2]==0x29 && ptr[3]==0x43 )
		return IsoKR;
	if( Cn && ptr[0]==0x1b && ptr[1]==24 && ptr[2]==0x29 && ptr[3]==0x41 )
		return IsoCN;
	if( Jp && !bit8 && freq[0x1b]>0 )
		return IsoJP;

//-- UTF-5 `FbN

	ulong u5sum = 0;
	for( uchar c='0'; c<='9'; ++c ) u5sum += freq[c];
	for( uchar c='A'; c<='V'; ++c ) u5sum += freq[c];
	if( siz == u5sum )
		return UTF5;

//-- b UTF-8 / {EUC `FbN

	// sR[hLFAx̑傫AłȂ
	//  ANSI-CP ƌȂĂ܂B
	if( bit8 && (siz>2048 || lb_==1) )
	{
		if( Jp )
		{
			// EUCƂĂl`FbN
			bool be=true;
			for( int k=0x90; k<=0xa0; ++k )if( freq[k]>0 ){be=false;break;}
			for( int k=0x7f; k<=0x8d; ++k )if( freq[k]>0 ){be=false;break;}
			if( be )
				return EucJP;
		}
		{
			// UTF8Ƃēǂ߂邩ǂ`FbN
			bool b8=true;
			int mi=1;
			for( ulong i=0; i<siz && b8; ++i )
				if( --mi )
				{
					if( ptr[i]<=0x80 || ptr[i]>=0xc0 )
						b8 = false;
				}
				else if( ptr[i] > 0x7f )
				{
					if( ptr[i] >= 0xfe )
						b8 = false;
					mi = GetMaskIndex( ptr[i] );
				}
			if( b8 )
				return UTF8;
		}
	}

//-- 茋

	return cs;
}
